/**
 * (C) Copyright IBM Corp. 2006, 2009
 *
 * THIS FILE IS PROVIDED UNDER THE TERMS OF THE ECLIPSE PUBLIC LICENSE 
 * ("AGREEMENT"). ANY USE, REPRODUCTION OR DISTRIBUTION OF THIS FILE 
 * CONSTITUTES RECIPIENTS ACCEPTANCE OF THE AGREEMENT.
 *
 * You can obtain a current copy of the Eclipse Public License from
 * http://www.opensource.org/licenses/eclipse-1.0.php
 *
 * @author : Alexander Wolf-Reber, IBM, a.wolf-reber@de.ibm.com
 * 
 * Change History
 * Flag       Date        Prog         Description
 *------------------------------------------------------------------------------- 
 * 1565892    2006-11-14  lupusalex    Make SBLIM client JSR48 compliant
 * 1711092    2006-05-02  lupusalex    Some fixes/additions of log&trace messages
 * 2003590    2008-06-30  blaschke-oss Change licensing from CPL to EPL
 * 2204488 	  2008-10-28  raman_arora  Fix code to remove compiler warnings
 * 2524131    2009-01-21  raman_arora  Upgrade client to JDK 1.5 (Phase 1)
 * 2531371    2009-02-10  raman_arora  Upgrade client to JDK 1.5 (Phase 2)
 * 2763216    2009-04-14  blaschke-oss Code cleanup: visible spelling/grammar errors
 */

package org.sblim.cimclient.internal.logging;

import java.io.IOException;
import java.io.OutputStream;
import java.text.MessageFormat;
import java.util.ArrayList;
import java.util.List;
import java.util.Properties;
import java.util.logging.ConsoleHandler;
import java.util.logging.FileHandler;
import java.util.logging.Handler;
import java.util.logging.Level;
import java.util.logging.LogRecord;
import java.util.logging.Logger;

import org.sblim.cimclient.LogAndTraceManager;
import org.sblim.cimclient.LogListener;
import org.sblim.cimclient.TraceListener;
import org.sblim.cimclient.internal.util.WBEMConfiguration;
import org.sblim.cimclient.internal.util.WBEMConstants;

/**
 * Class LogAndTraceBroker is the central class that implements the logging and
 * tracing of the CIM Client. It manages the collections of the internal log and
 * trace listeners. It sets up the application independent logging. It provides
 * the API to send log and trace messages and forwards them to the appropriate
 * listeners.
 * 
 */
public class LogAndTraceBroker {

	private static final String TRACE_LOGGER = "org.sblim.cimclient.trace";

	private static final String FILE_LOGGER = "org.sblim.cimclient.file";

	private static final String CONSOLE_LOGGER = "org.sblim.cimclient.console";

	private static LogAndTraceBroker cBroker = null;

	/**
	 * Returns the singleton instance of the broker
	 * 
	 * @return The broker instance
	 */
	public static LogAndTraceBroker getBroker() {
		if (cBroker == null) {
			synchronized (LogAndTraceBroker.class) {
				if (cBroker == null) {
					cBroker = new LogAndTraceBroker();
				}
			}
		}
		return cBroker;
	}

	/**
	 * Returns if the logging framework has been initialized. This method is
	 * used by the <code>WBEMConfiguration</code> class to determine if the
	 * logging is already up. The <code>WBEMConfiguration</code> is initialized
	 * before the logging, so methods in this class cannot assume the logging to
	 * be up and running.
	 * 
	 * @return <code>true</code> if the logging is up, <code>false</code>
	 *         otherwise
	 */
	public static boolean isLoggingStarted() {
		return cBroker != null;
	}

	private volatile ArrayList<LogListener> iLogListeners;

	private volatile ArrayList<TraceListener> iTraceListeners;

	private String iProductName = "SBLIM CIM Client for Java";

	private String iCopyright = "COPYRIGHT (C) 2006, 2009 IBM Corp.";

	private String iVersion = "?";

	private String iBuildDate = "?";

	private String iBuildTime = "?";

	private OutputStream iXmlTraceStream = null;

	private final String iTHIS_CLASS = this.getClass().getName();

	private final String iTHROWABLE = Throwable.class.getName();

	private LogAndTraceBroker() {
		this.iLogListeners = new ArrayList<LogListener>();
		this.iTraceListeners = new ArrayList<TraceListener>();
		loadVersionTxt();
		registerInternalListeners();
	}

	/**
	 * Registers the listeners for our internal loggers
	 */
	public void registerInternalListeners() {
		try {
			Level level = WBEMConfiguration.getGlobalConfiguration().getLogConsoleLevel();
			String type = WBEMConfiguration.getGlobalConfiguration().getLogConsoleType();
			if (level.intValue() < Level.OFF.intValue() && WBEMConstants.MESSAGE.equals(type)) {
				final Logger logger = Logger.getLogger(CONSOLE_LOGGER);
				final Handler handler = new ConsoleHandler();
				handler.setFormatter(new LogFormatter());
				handler.setLevel(level);
				logger.addHandler(handler);
				logger.setLevel(level);
				logger.setUseParentHandlers(false);
				addLogListener(new LogListener() {

					public void log(Level pLevel, String pMessageKey, String pMessage,
							Object[] pParameters) {
						LogRecord record = new LogRecord(pLevel, pMessageKey + " " + pMessage);
						record.setParameters(pParameters);
						logger.log(record);
					}
				});
			}
		} catch (Exception e) {
			// Don't crash for logging
		}
		try {
			Level level = WBEMConfiguration.getGlobalConfiguration().getLogFileLevel();
			String location = WBEMConfiguration.getGlobalConfiguration().getLogFileLocation();
			int size = WBEMConfiguration.getGlobalConfiguration().getLogFileSizeLimit();
			int count = WBEMConfiguration.getGlobalConfiguration().getLogFileCount();
			if (level.intValue() < Level.OFF.intValue()) {
				final Logger logger = Logger.getLogger(FILE_LOGGER);
				final Handler handler = new FileHandler(location, size, count);
				handler.setFormatter(new LogFormatter());
				handler.setLevel(level);
				logger.addHandler(handler);
				logger.setLevel(level);
				logger.setUseParentHandlers(false);
				addLogListener(new LogListener() {

					public void log(Level pLevel, String pMessageKey, String pMessage,
							Object[] pParameters) {
						LogRecord record = new LogRecord(pLevel, pMessageKey + " " + pMessage);
						record.setParameters(pParameters);
						logger.log(record);
					}
				});
			}
		} catch (Exception e) {
			// Don't crash for logging
		}
		try {
			Level level = WBEMConfiguration.getGlobalConfiguration().getLogConsoleLevel();
			String type = WBEMConfiguration.getGlobalConfiguration().getLogConsoleType();
			if (level.intValue() < Level.OFF.intValue() && WBEMConstants.TRACE.equals(type)) {
				final Logger logger = Logger.getLogger(CONSOLE_LOGGER);
				final Handler handler = new ConsoleHandler();
				handler.setFormatter(new TraceFormatter());
				handler.setLevel(level);
				logger.addHandler(handler);
				logger.setLevel(level);
				logger.setUseParentHandlers(false);
				addTraceListener(new TraceListener() {

					public void trace(Level pLevel, StackTraceElement pOrigin, String pMessage) {
						LogRecord record = new LogRecord(pLevel, pMessage);
						record.setSourceMethodName(String.valueOf(pOrigin));
						logger.log(record);
					}

					public void trace(Level pLevel, StackTraceElement pOrigin, String pMessage,
							Throwable pThrown) {
						LogRecord record = new LogRecord(pLevel, pMessage);
						record.setSourceMethodName(String.valueOf(pOrigin));
						record.setThrown(pThrown);
						logger.log(record);
					}
				});
			}
		} catch (Exception e) {
			// Don't crash for logging
		}
		try {
			Level level = WBEMConfiguration.getGlobalConfiguration().getTraceFileLevel();
			String location = WBEMConfiguration.getGlobalConfiguration().getTraceFileLocation();
			int size = WBEMConfiguration.getGlobalConfiguration().getTraceFileSizeLimit();
			int count = WBEMConfiguration.getGlobalConfiguration().getTraceFileCount();
			if (level.intValue() < Level.OFF.intValue()) {
				final Logger logger = Logger.getLogger(TRACE_LOGGER);
				final Handler handler = new FileHandler(location, size, count);
				handler.setFormatter(new TraceFormatter());
				handler.setLevel(level);
				logger.addHandler(handler);
				logger.setLevel(level);
				logger.setUseParentHandlers(false);
				addTraceListener(new TraceListener() {

					public void trace(Level pLevel, StackTraceElement pOrigin, String pMessage) {
						LogRecord record = new LogRecord(pLevel, pMessage);
						record.setSourceMethodName(String.valueOf(pOrigin));
						logger.log(record);
					}

					public void trace(Level pLevel, StackTraceElement pOrigin, String pMessage,
							Throwable pThrown) {
						LogRecord record = new LogRecord(pLevel, pMessage);
						record.setSourceMethodName(String.valueOf(pOrigin));
						record.setThrown(pThrown);
						logger.log(record);
					}
				});
			}
		} catch (Exception e) {
			// Don't crash for logging
		}
	}

	/**
	 * Adds a listener for log messages. The listener will be notified of any
	 * log event. Uses copy on write to ensure concurrent read access.
	 * 
	 * @param pListener
	 *            The listener
	 */
	public synchronized void addLogListener(LogListener pListener) {
		sendGreetings(pListener);
		ArrayList<LogListener> newListeners = new ArrayList<LogListener>(this.iLogListeners);
		newListeners.add(pListener);
		this.iLogListeners = newListeners;
	}

	private void loadVersionTxt() {
		try {
			Properties version = new Properties();
			version.load(LogAndTraceManager.class.getResourceAsStream("version.txt"));
			this.iProductName = version.getProperty("PRODUCTNAME");
			this.iCopyright = version.getProperty("COPYRIGHT");
			this.iVersion = version.getProperty("VERSION");
			this.iBuildDate = version.getProperty("BUILDDATE");
			this.iBuildTime = version.getProperty("BUILDTIME");
		} catch (Exception e) {
			// nothing to do
		}
	}

	private void sendGreetings(LogListener pListener) {

		pListener.log(MessageLoader.getLevel(Messages.GREETING), Messages.GREETING, MessageLoader
				.getLocalizedMessage(Messages.GREETING), new Object[] { this.iProductName,
				this.iCopyright });
		pListener.log(MessageLoader.getLevel(Messages.RELEASE), Messages.RELEASE, MessageLoader
				.getLocalizedMessage(Messages.RELEASE), new Object[] { this.iVersion,
				this.iBuildDate, this.iBuildTime });
		pListener.log(MessageLoader.getLevel(Messages.OS), Messages.OS, MessageLoader
				.getLocalizedMessage(Messages.OS), new Object[] { System.getProperty("os.name"),
				System.getProperty("os.version"), System.getProperty("os.arch") });
		pListener.log(MessageLoader.getLevel(Messages.JRE), Messages.JRE, MessageLoader
				.getLocalizedMessage(Messages.JRE), new Object[] {
				System.getProperty("java.version"), System.getProperty("java.vendor") });
		pListener.log(MessageLoader.getLevel(Messages.JVM), Messages.JVM, MessageLoader
				.getLocalizedMessage(Messages.JVM), new Object[] {
				System.getProperty("java.vm.name"), System.getProperty("java.vm.version"),
				System.getProperty("java.vm.vendor") });
		pListener.log(MessageLoader.getLevel(Messages.CONFIGURATION_URL),
				Messages.CONFIGURATION_URL, MessageLoader
						.getLocalizedMessage(Messages.CONFIGURATION_URL),
				new Object[] { WBEMConfiguration.getActiveConfigURL() });
		if (!WBEMConfiguration.isConfigurationLoadSuccessful()) {
			pListener.log(MessageLoader.getLevel(Messages.CONFIGURATION_LOAD_FAILED),
					Messages.CONFIGURATION_LOAD_FAILED, MessageLoader
							.getLocalizedMessage(Messages.CONFIGURATION_LOAD_FAILED),
					(Object[]) null);
			if (WBEMConfiguration.getConfigurationLoadException() != null) {
				pListener.log(MessageLoader.getLevel(Messages.EXCEPTION_DURING_CONFIGURATION_LOAD),
						Messages.EXCEPTION_DURING_CONFIGURATION_LOAD, MessageLoader
								.getLocalizedMessage(Messages.EXCEPTION_DURING_CONFIGURATION_LOAD),
						new Object[] { WBEMConfiguration.getConfigurationLoadException()
								.getMessage() });
			}
		}
	}

	private void sendGreetings(TraceListener pListener) {
		StackTraceElement origin = new Throwable().getStackTrace()[0];
		pListener.trace(MessageLoader.getLevel(Messages.GREETING), origin, MessageFormat.format(
				MessageLoader.getMessage(Messages.GREETING), new Object[] { this.iProductName,
						this.iCopyright }));
		pListener.trace(MessageLoader.getLevel(Messages.RELEASE), origin, MessageFormat.format(
				MessageLoader.getMessage(Messages.RELEASE), new Object[] { this.iVersion,
						this.iBuildDate, this.iBuildTime }));
		pListener.trace(MessageLoader.getLevel(Messages.OS), origin, MessageFormat.format(
				MessageLoader.getMessage(Messages.OS), new Object[] {
						System.getProperty("os.name"), System.getProperty("os.version"),
						System.getProperty("os.arch") }));
		pListener.trace(MessageLoader.getLevel(Messages.JRE), origin, MessageFormat.format(
				MessageLoader.getMessage(Messages.JRE), new Object[] {
						System.getProperty("java.version"), System.getProperty("java.vendor") }));
		pListener.trace(MessageLoader.getLevel(Messages.JVM), origin, MessageFormat.format(
				MessageLoader.getMessage(Messages.JVM), new Object[] {
						System.getProperty("java.vm.name"), System.getProperty("java.vm.version"),
						System.getProperty("java.vm.vendor") }));
		pListener.trace(MessageLoader.getLevel(Messages.CONFIGURATION_URL), origin, MessageFormat
				.format(MessageLoader.getMessage(Messages.CONFIGURATION_URL),
						new Object[] { WBEMConfiguration.getActiveConfigURL() }));
		if (!WBEMConfiguration.isConfigurationLoadSuccessful()) {
			pListener.trace(MessageLoader.getLevel(Messages.CONFIGURATION_LOAD_FAILED), origin,
					MessageLoader.getMessage(Messages.CONFIGURATION_LOAD_FAILED));
			if (WBEMConfiguration.getConfigurationLoadException() != null) {
				pListener.trace(MessageLoader
						.getLevel(Messages.EXCEPTION_DURING_CONFIGURATION_LOAD), origin,
						MessageLoader.getMessage(Messages.EXCEPTION_DURING_CONFIGURATION_LOAD),
						WBEMConfiguration.getConfigurationLoadException());
			}
		}
	}

	/**
	 * Remove a listener. This listener will not be notified of log events
	 * anymore.
	 * 
	 * @param pListener
	 *            The listener
	 */
	public synchronized void removeLogListener(LogListener pListener) {
		ArrayList<LogListener> newListeners = new ArrayList<LogListener>(this.iLogListeners);
		newListeners.remove(pListener);
		this.iLogListeners = newListeners;
	}

	/**
	 * Removes all listeners. Caution this will also remove the internal console
	 * and file loggers.
	 */
	public synchronized void clearLogListeners() {
		this.iLogListeners = new ArrayList<LogListener>();
		removeHandlers(Logger.getLogger(CONSOLE_LOGGER));
		removeHandlers(Logger.getLogger(FILE_LOGGER));
	}

	/**
	 * Gets the registered log listeners including the internal console and file
	 * loggers.
	 * 
	 * @return The list of listeners
	 */
	public List<LogListener> getLogListeners() {
		return this.iLogListeners;
	}

	/**
	 * Adds a listener for log messages. The listener will be notified of any
	 * trace event.
	 * 
	 * @param pListener
	 *            The listener
	 */
	public synchronized void addTraceListener(TraceListener pListener) {
		sendGreetings(pListener);
		ArrayList<TraceListener> newListeners = new ArrayList<TraceListener>(this.iTraceListeners);
		newListeners.add(pListener);
		this.iTraceListeners = newListeners;
	}

	/**
	 * Removes a listener. This listener will not be notified of trace events
	 * anymore.
	 * 
	 * @param pListener
	 *            The listener
	 */
	public synchronized void removeTraceListener(TraceListener pListener) {
		ArrayList<TraceListener> newListeners = new ArrayList<TraceListener>(this.iTraceListeners);
		newListeners.remove(pListener);
		this.iTraceListeners = newListeners;
	}

	/**
	 * Removes all listeners. Caution this will also remove the internal trace
	 * file listener.
	 */
	public synchronized void clearTraceListeners() {
		this.iTraceListeners = new ArrayList<TraceListener>();
		removeHandlers(Logger.getLogger(TRACE_LOGGER));
	}

	/**
	 * Gets the registered trace listeners including the internal console and
	 * file loggers.
	 * 
	 * @return A list of listeners
	 */
	public List<TraceListener> getTraceListeners() {
		return this.iTraceListeners;
	}

	/**
	 * Forwards a log/trace message to the registered log&trace listeners.
	 * 
	 * @param pKey
	 *            The message identifier.
	 */
	public void message(String pKey) {
		message(pKey, (Object[]) null);
	}

	/**
	 * Forwards a log/trace message to the registered log&trace listeners.
	 * 
	 * @param pKey
	 *            The message identifier.
	 * @param pParameter
	 *            The parameter for the message
	 */
	public void message(String pKey, Object pParameter) {
		message(pKey, new Object[] { pParameter });
	}

	/**
	 * Forwards a log/trace message to the registered log&trace listeners.
	 * 
	 * @param pKey
	 *            The message identifier.
	 * @param pParameters
	 *            The parameters for the message
	 */
	public void message(String pKey, Object[] pParameters) {
		try {
			final String message = MessageLoader.getMessage(pKey);
			final String localMessage = MessageLoader.getLocalizedMessage(pKey);
			final Level level = MessageLoader.getLevel(pKey);
			final List<TraceListener> traceListeners = getTraceListeners();
			StackTraceElement caller = null;
			for (int i = 0; i < traceListeners.size(); ++i) {
				caller = caller == null ? getCaller() : caller;
				traceListeners.get(i).trace(level, caller,
						pKey + " " + MessageFormat.format(message, pParameters));
			}
			final List<LogListener> logListeners = getLogListeners();
			for (int i = 0; i < logListeners.size(); ++i) {
				logListeners.get(i).log(level, pKey, localMessage, pParameters);
			}
		} catch (Exception e) {
			// don't crash for logging
		}
	}

	/**
	 * Forwards a trace message to the registered trace listeners.
	 * 
	 * @param pLevel
	 *            One of the three message level identifiers FINE, FINER and
	 *            FINEST
	 * @param pMessage
	 *            The message text
	 */
	public void trace(Level pLevel, String pMessage) {
		try {
			final List<TraceListener> traceListeners = getTraceListeners();
			StackTraceElement caller = null;
			for (int i = 0; i < traceListeners.size(); ++i) {
				caller = caller == null ? getCaller() : caller;
				traceListeners.get(i).trace(pLevel, caller, pMessage);
			}
		} catch (Exception e) {
			// don't crash for logging
		}
	}

	/**
	 * Forwards a trace message to the registered trace listeners.
	 * 
	 * @param pLevel
	 *            One of the three message level identifiers FINE, FINER and
	 *            FINEST
	 * @param pMessage
	 *            The message text
	 * @param pThrown
	 *            The throwable associated with the message
	 */
	public void trace(Level pLevel, String pMessage, Throwable pThrown) {
		try {
			final List<TraceListener> traceListeners = getTraceListeners();
			StackTraceElement caller = null;
			for (int i = 0; i < traceListeners.size(); ++i) {
				caller = caller == null ? getCaller() : caller;
				traceListeners.get(i).trace(pLevel, caller, pMessage, pThrown);
			}
		} catch (Exception e) {
			// don't crash for logging
		}
	}

	/**
	 * Forwards a method entry message to the registered trace listeners.
	 */
	public void entry() {
		trace(Level.FINEST, "Entering method");
	}

	/**
	 * Forwards a method exit message to the registered trace listeners.
	 */
	public void exit() {
		trace(Level.FINEST, "Exiting method");
	}

	/**
	 * Returns the output stream to which all CIM-XML traffic (outgoing &amp;
	 * incoming) will be copied for debugging purposes.
	 * 
	 * @return The output stream. A <code>null</code> value means that CIM-XML
	 *         debugging is disabled
	 */
	public OutputStream getXmlTraceStream() {
		return this.iXmlTraceStream;
	}

	/**
	 * Sets an output stream to which all CIM-XML traffic (outgoing &amp;
	 * incoming) will be copied for debugging purposes.
	 * 
	 * @param pStream
	 *            The output stream. A <code>null</code> value means that
	 *            CIM-XML debugging is disabled.
	 */
	public void setXmlTraceStream(OutputStream pStream) {
		if (this.iXmlTraceStream != null) {
			try {
				this.iXmlTraceStream.close();
			} catch (IOException e) {
				// just ignore it
			}
		}
		this.iXmlTraceStream = pStream;
	}

	/**
	 * Analyzes the stack trace and determines from where the
	 * <code>LogAndTraceBroker</code> was called.
	 * 
	 * @return First <code>StackTraceElement</code> outside the
	 *         <code>LogAndTraceBroker</code>
	 */
	private StackTraceElement getCaller() {
		StackTraceElement[] stack = (new Throwable()).getStackTrace();
		for (int i = 0; i < stack.length; ++i) {
			StackTraceElement frame = stack[i];
			String cname = frame.getClassName();
			if (!this.iTHIS_CLASS.equals(cname) && !this.iTHROWABLE.equals(cname)) { return frame; }
		}
		return null;
	}

	/**
	 * Removes all handlers from a logger
	 * 
	 * @param pLogger
	 *            The logger
	 */
	private void removeHandlers(Logger pLogger) {
		Handler[] handlers = pLogger.getHandlers();
		for (int i = 0; i < handlers.length; ++i) {
			pLogger.removeHandler(handlers[i]);
			handlers[i].close();
		}
	}

}
